#include <cstdlib>
#include "utils.PRO2"
#include "Arbre.hpp"
#include "ArbIOint.hpp"


//Un arbre eś balancejat si totes les seves branques tenen com a màxim
//un de diferència en llargada


/* Pre: a=A */
pair<int,int> minmaxaltura(Arbre<int>& a)
{
  if (a.es_buit()) return make_pair(0,0);
  else{
    Arbre<int> a1;
    Arbre<int> a2;
    a.fills(a1,a2);
    pair<int,int> y=minmaxaltura(a1);
    pair<int,int> z=minmaxaltura(a2);
    return make_pair(min(y.first,z.first)+1,max(y.second,z.second)+1);
  }
}
/* Post: El resultat són respectivament les longitud de la branca més curta
   i la més llarga de l'arbre A */


//1- Solució ineficient

/* Pre: a=A */
int balancejat(Arbre<int>& a){
  if (a.es_buit()) return true;
  else{
    Arbre<int> a1;
    Arbre<int> a2;
    a.fills(a1,a2);
    Arbre<int> b1=a1;
    Arbre<int> b2=a2;
    pair<int,int> y=minmaxaltura(b1);
    pair<int,int> z=minmaxaltura(b2);
    if (abs(z.second-y.first)>1 or abs(y.second-z.first)>1) return false;
    else if (!balancejat(a1)) return false;
    else return balancejat(a2);
  }
}
/* Post: si l'arbre A és balancejat*/


//2- Primera solució eficient

/* Pre: a=A */
pair<bool,pair<int,int> > ebalancejat(Arbre<int>& a){
  if (a.es_buit()) return make_pair(true,make_pair(0,0));
  else{
    Arbre<int> a1;
    Arbre<int> a2;
    a.fills(a1,a2);
    pair<bool,pair<int,int> > ba1=ebalancejat(a1);
    //HI1: retorna si el fill esq d'A és balancejat i la mínima i la
    //     màxima longitut de les branques del fill esq d'A
    pair<bool,pair<int,int> > ba2=ebalancejat(a2);
    //HI2: retorna si el fill dre d'A és balancejat i la mínima i la
    //     màxima longitut de les branques del fill dret d'A
    int min1 = ba1.second.first;
    int max1 = ba1.second.second;
    int min2 = ba2.second.first;
    int max2 = ba2.second.second;
    bool ba = abs(max1-min2)<=1 and abs(max2-min1)<=1 and ba1.first and ba2.first;
    return make_pair(ba,make_pair(min(min1,min2)+1,max(max1,max2)+1));
  }
}
/* Post: retorna a first si l'arbre A és balancejat i 
   a second la mínima i la màxima longitud de les branques d'A*/


//2- Segona solució eficient amb postcondició condicionada. És a dir
//només cal retornar la mínima i la màxima longitud si l'abre és balancejat


/* Pre: a=A */
pair<bool,pair<int,int> > ebalancejat1(Arbre<int>& a){
  if (a.es_buit()) return make_pair(true,make_pair(0,0));
  else{
    Arbre<int> a1;
    Arbre<int> a2;
    a.fills(a1,a2);
    pair<bool,pair<int,int> > ba1=ebalancejat1(a1);
    //HI1: retorna si el fill esq d'A és balancejat i la mínima i la
    //     màxima longitut de les branques del fill esq d'A
    if (!ba1.first) return ba1;
    pair<bool,pair<int,int> > ba2=ebalancejat1(a2);
    //HI2: retorna si el fill dre d'A és balancejat i la mínima i la
    //     màxima longitut de les branques del fill dret d'A
    if (!ba2.first) return ba2;
    int min1 = ba1.second.first;
    int max1 = ba1.second.second;
    int min2 = ba2.second.first;
    int max2 = ba2.second.second;
    bool ba = abs(max1-min2)<=1 and abs(max2-min1)<=1;
    return make_pair(ba,make_pair(min(min1,min2)+1,max(max1,max2)+1));
  }
}
/* Post: retorna a first si l'arbre A és balancejat i si ho és 
         retorna a second la mínima i la màxima longitud de les branques d'A*/



int main(){

  Arbre<int> a;
  llegir_arbre_int(a,0);
  Arbre<int> b=a;
  if (balancejat(a)) cout << "És balancejat"<< endl;
  else cout << "No és balancejat"<< endl;
  a=b;
  if (ebalancejat(b).first) cout << "És balancejat"<< endl;
  else cout << "No és balancejat"<< endl;
  if (ebalancejat1(a).first) cout << "És balancejat"<< endl;
  else cout << "No és balancejat"<< endl;
}


/* Justificació de "ebalancejat1"

1) Cas senzill.

Si A és buit les branques tenen longitud 0 i per tant és balancejat i
la mínima i la màxima longitud de les branques d'A és 0.

2) Cas resursiu.

Les crides són correctes. Com que no és buit, A té fill esquerra i
fill dret. Per inducció (HI1), ba1.first ens diu si el fill esquerra
d'A és balancejat o no. Si el fill esquerra no és balancejat llavors A
no és balancejat, i per tant podem tornar el mateix ba1. Si és
balancejat llavors a ba1.second tenim la mínima i la màxima longitud de
les branques del fill esquerra d'A. Per inducció (HI2), ba2.first ens
diu si el fill dret d'A és balancejat o no. Si el fill dret no és
balancejat llavors A no és balancejat, i per tant podem tornar el
mateix ba2. Si és balancejat llavors ba2.second tenim la mínima i la
màxima longitud de les branques del fill dret d'A.

Si els dos fills són balancejats, llavors A és balancejat si i només
si la distància entre la mínima branca de cada fill i la màxima de
l'altre és un. Finalment, la mínima i la màxima longitutd de les
branques d'A són respectivament igual al mínim de les mínimes i al
màxim de les màximes dels fills més 1.

3) Acabament.

Decreix la mida d'a, ja que tant a1 com a2 tenen com a mínim un node
menys.


 */
