#include "utils.PRO2"
#include "Arbre.hpp"
#include "ArbIOint.hpp"


/* Un arbre és complet si tots els sub arbres a la mateixa profunditat
   tenen els mateixos nodes */

/* Pre: a=A */
int mida(Arbre<int>& a)
{
  int n;
  if (a.es_buit()) n=0;
  else{
    Arbre<int> a1;
    Arbre<int> a2;
    a.fills(a1,a2);
    int n1=mida(a1);
    //{HI1: n1= "mida del fill esquerra d'A" }
    int n2=mida(a2);
    //{HI2: n2= "mida del fill dret d'A" }
    n = n1 + n2 + 1;
  }
/* n= "mida d'A" */
  return n;
}
/* Post: retorna la mida d'A */



//Versió ineficient. L'algorisme fa de l'ordre de n^2 operacions (on n
//és la mida de l'arbre)

/* Pre: a=A */
bool complet(Arbre<int>& a){
  bool b;
  if (a.es_buit()) b=true;
  else{
    Arbre<int> a1;
    Arbre<int> a2;
    a.fills(a1,a2);
    Arbre<int> b1=a1;
    Arbre<int> b2=a2;
    int n1=mida(b1);
    int n2=mida(b2);
    if (n1!=n2) b=false;
    else if (not complet(a1)) //HI1a: fill esquerra d'A no és complet
      b=false;
    else //HI1b: fill esquerra d'A és complet
      b= complet(a2); //HI2: b="fill dret d'A és complet"
  }
/* Post1: b= "A és complet" */
  return b;
}  
/* Post: retorna si A és complet o no*/




/* 

Justificació:

1) Cas senzill.

Si A és buit no té cap fill i per tant és complet.

2) Cas resursiu.

Les crides són correctes. Com que no és buit, A té fill esquerra i
fill dret. Així, si el fill esquerra i el fill dret no tenen la mateix
mida tenim que A no és complet. Si tenen la mateix mida si complet(a1)
es fals llavors per inducció tenim HI1a és a dir el fill esquerra d'A
no és complet i per tant A no és complet. Sino, per inducció tenim
HI1b, és a dir, que el fill esquerra d'A és complet. En aquest cas A
és complet si i només si ho és el seu fill dret, i per inducció això
és el que tenim a b.

3) Acabament.

Decreix la mida d'a, ja que tant a1 com a2 tenen com a mínim un node
menys.

*/


//Versió eficient. L'algorisme fa de l'ordre de n operacions (on n
//és la mida de l'arbre)


/* Pre: a=A */
pair<bool,int> ecomplet(Arbre<int>& a){
  bool b;
  int n;
  if (a.es_buit()) {
        b=true;
        n=0;
  }
  else{
    Arbre<int> a1;
    Arbre<int> a2;
    a.fills(a1,a2);
    pair<bool,int> p1 = ecomplet(a1);
    //HI1: p1.first="fill esquerra d'A és complet"; 
    //     si p1.first llavors p1.second= "mida del fill esquerra d'A"
    if (not p1.first) b = false;
    else {
      pair<bool,int> p2 = ecomplet(a2);
      //HI2: p2.first="fill dret d'A és complet"; 
      //si p2.first llavors p2.second= "mida del fill dret d'A"
      if (not p2.first) b=false;
      else {
        b=(p1.second==p2.second);
        n=p1.second+p2.second+1;
      }
    }
  }
  /* Post1: b= "A és complet"; Si b llavors n = "mida d'A" */
  return make_pair(b,n);
}  
//Post: retorna un parel on el first="fill dret d'A és complet" i 
//      Si first és cert llavors el second= "mida del fill dret d'A"


/* 

Justificació:

1) Cas senzill.

Si A és buit no té cap fill i per tant és complet i la mida és 0.

2) Cas resursiu.

Les crides són correctes. Com que no és buit, A té fill esquerra i
fill dret. Per inducció (HI1), p1.first ens diu si el fill esquerra
d'A és complet o no. Si el fill esquerra no és complet llavors A no és
complet (i no cal donar valor a n). Si és complet llavors
p1.second="mida del fill esquerra d'A". Per inducció (HI2), p2.first
ens diu si el fill dret d'A és complet o no. Si el fill dret
no és complet llavors A no és complet (i no cal donar valor a n). Si és
complet llavors p2.second="mida del fill dret d'A". 

Si els dos fills són complets, llavors A és complet si i només si
tenen la mateix mida, és a dir, si p1.second = p2.second. Finalment,
la mida d'A és igual a la mida dels fills més 1.

3) Acabament.

Decreix la mida d'a, ja que tant a1 com a2 tenen com a mínim un node
menys.

*/


int main(){
  Arbre<int> a1;
  Arbre<int> a2;
  llegir_arbre_int(a1,0);
  a2=a1;
  bool b = complet(a1);
  if (b) cout << "És complet" << endl;
  else cout << "No és complet" << endl;
  pair<bool,int> p = ecomplet(a2);
  if (p.first) cout << "És complet" << endl;
  else cout << "No és complet" << endl;  
}
