#include "../includes/utils.PRO2"
#include <vector>


//Funcions recursives amb vectors. Només com exemple de com fer
//funcions recursives sobre vectors. Per tots aquests exemples la
//versió iterativa sería la raonable (especialment pels de suma del
//vectors).
//De les versions suma només hi ha la justificació de la primera

//Pre: i <= v.size()-1
int sumaVecRec(const vector<int> & v, int i){
int s;
if (i<0) s=0;
else { // i>=0
  s=sumaVecRec(v,i-1);
  // HI: s = "suma de v[0..i-1]"
  s+=v[i];
}
//Post1: s= "suma de v[0..i]"
return s;
}
//Post: retorna la suma de v[0..i]


/*

Justificació:

Crida inicial: veure main.  
Com que hem afegit paràmetres cal posar-la i hem canviat la
postcondició, cal comprovar que es satisfà la precondició i que el
resultat satisfà la precondició original que era {s = "suma de v"}

         sumaVecRec(v,v.size()-1)

La precondició es satisfà ja que v.size()-1 <= v.size()-1
La funció retorna la suma de v[0..v.size()-1], és a dir la suma de v.

Cas senzill: 
Si i<0 llavors l'interval [0..i] és buit i per tant la suma és 0.

Cas recursiu: 
La cirda és correcta ja que i<= v.size()-1 garanteix i-1<= v.size()-1;
Per inducció (HI) tenim que s = "suma de v[0..i-1]" per tant s+v[i] és
la suma de v[0..i].

Acabament: decreix i. Noteu que de la Pre i el cas recursiu tenim
que i>=0.

 */


//Pre: i>=0
int sumaVecRec2(const vector<int> & v, int i){
int s;
if (i>= v.size()) s=0;
else{
  s= sumaVecRec2(v,i+1);
  //H.I.: s="suma de v[i+1..v.size()-1]"
  s+=v[i];
}
//Post1: s= "suma de v[i..v.size()-1]"
return s;
}
//Post: retorna la suma de v[i..v.size()-1]



//Pre: s= suma de v[0..i-1], 0 =< i =< v.size()
int sumaVecRec3(const vector<int> & v, int i, int & s){
  if (i==v.size()) return s;
  else return sumaVecRec3(v, i+1, s+v[i]);
}
//Post: retorna la suma de v


//Cerca dicotòmica. Es pot fer també iterativa.

//Pre: 0 <= i; j < v.size(); v està ordenat creixent
bool cercaDicRec(const vector<int> & v, int x, int i, int j){
  bool b;
  if (i>j) b= false;
  else {
    int k = (i+j)/2;
    if (v[k] == x) b= true;
    else if (v[k]<x) 
      b= cercaDicRec(v,x,k+1,j);
      //HI1: b = "x pertany a v[k+1..j]"
    else // x< v[k]
     b= cercaDicRec(v,x,i,k-1);
     //HI2: b= "x pertany a v[i..k-1]"
  }
  //Post1: b = "x pertany a v[i..j]"
  return b;
}
//Post: retorna si x pertany o no a v[i..j]

/*

Justificació:

Crida inicial: veure main.  

         cercaDicRec(v,x,0,v.size()-1))

La precondició es satisfà ja que 0 <= 0 i v.size()-1 < v.size()
La funció retorna si x pertany o no a v[0..v.size()-1], és a dir a v.

Casos senzills:

Si i>j llavors l'interval [i..j] és buit i per tant x no pertany.

Si i<=j llavors, com que k = (i+j)/2, tenim que i<=k<=j. Per tant si
v[k]=x llavors x pertany a v[i..j].

Casos recursius:

Com que tenim i<=k<=j, hi ha dos casos:

Si v[k]<x, com el vector està ordenat sabem que tots els valors de
v[i..k] són més petits que x i per tant x no pertany a v[i..k]. Això
implica que x pertany a v[i..j] si i només si x pertany a v[k+1..j],
que és el que obtenim per inducció (HI1) a b.

Si v[k]>x, com el vector està ordenat sabem que tots els valors de
v[k..j] són més grans que x i per tant x no pertany a v[k..j]. Això
implica que x pertany a v[i..j] si i només si x pertany a v[i..k-1],
que és el que obtenim per inducció (HI2) a b.

Acabament: decreix j-i, és a dir, la distància entre els dos.
De la precondició i el casos recursius (i<=j) tenim que j-i>=0.
A més la distancància decreix ja que k-1 < j  i  k+1 > i.

*/


//Funcions recursives amb matrius. Només com exemple de com fer
//funcions recursives sobre matrius. Per tots aquests exemples la
//versió iterativa sería la raonable.

//Pre: -1 =< i < m.size()
int zerosMatRec(const vector<vector<int> > & m, int i){
  if (i<0) return 0;
  else {
    int c=0;
    //c= "nombre de zeros de m[i][0..j-1]"
    for(unsigned int j=0; j< m[i].size(); ++j) 
      if (m[i][j]==0) ++c;

    return zerosMatRec(m, i-1) + c;
  }
}
//Post: retorna quants zeros hi ha a m[0..i]


//Pre: -1 =< i < m.size()
int zerosVecRec(const vector<int> & v, int i){
  if (i<0) return 0;
  else {
    int c=zerosVecRec(v, i-1);
    //HI: c= "nombre de zeros de v[0..i-1]"
    if (v[i]==0) ++c;
    return c;
  }
}
//Post: retorna quants zeros hi ha a v[0..i]



//Pre: -1 =< i < m.size()
int zerosMatRecRec(const vector<vector<int> > & m, int i){
  if (i<0) return 0;
  else {
    int c=zerosVecRec(m[i],m[i].size()-1);
    //c= "nombre de zeros de m[i]"
    int z= zerosMatRecRec(m, i-1) 
    //HI z= "quants zeros hi ha a m[0..i-1]"
    return z + c;
  }
}
//Post: retorna quants zeros hi ha a m[0..i]



void llegir_vector_int(vector<int> & v){
  for (unsigned int i=0; i<v.size(); ++i) cin >> v[i];
}

int main(){
  cout << "Tamany del vector: ";
  int n;
  cin >> n;
  vector<int> v(n);
  llegir_vector_int(v);
  cout << endl << "Suma: " << sumaVecRec(v,v.size()-1) << endl;
  cout << endl << "Suma: " << sumaVecRec2(v,0) << endl;
  cout << endl << "Suma: " << sumaVecRec3(v,0,0) << endl;
  cout << "Número a crecar: ";
  int x;
  cin >> x;
  if (cercaDicRec(v,x,0,v.size()-1)) cout << endl << "Hi és" << endl;
  else cout << endl << "No hi és" << endl;
  cout << "Files de la matriu: ";
  cin >> n;
  cout << endl << "Columnes de la matriu: ";
  int m;
  cin >> m;
  vector<vector<int> > mat(n,vector<int>(m));
  for (unsigned int i=0; i<mat.size(); ++i) 
    llegir_vector_int(mat[i]);
  cout << endl << "Nombre de zeros: " << zerosMatRec(mat,mat.size()-1)<< endl;
  cout << endl << "Nombre de zeros: " << zerosMatRecRec(mat,mat.size()-1)<< endl;
}

